import numpy as np
from scipy.stats import norm
import os
from multiprocess import Pool
#from pathos.multiprocessing import ProcessingPool as Pool

from .estimators import *

def add_column(self, name, values):
    self.data[name] = np.nan
    self.data.loc[self.active_index,name] = values
    
def make_ci_asy(self, confidence, hyp):
    
    one = norm.ppf(confidence/100)
    two = norm.ppf((confidence+(100-confidence)/2)/100)
    
    self.ci_one = np.sqrt(self.intKsq)*one
    self.ci_two = np.sqrt(self.intKsq)*two
    
    if hyp == 'twosided':
        self.ci = self.ci_two
    if hyp == 'onesided':
        self.ci = self.ci_one

    self.core_ci = self.ci*self.hat_q/np.sqrt(self.sample_size*self.band)
        
    add_column(self, '_q_ci_asy', self.core_ci)
    add_column(self, '_v_ci_asy', self.A_4*self.core_ci)
    add_column(self, '_bs_ci_asy', self.a*self.A_3*self.A_4*self.core_ci)
    add_column(self, '_rev_ci_asy', self.M*self.a*self.A_3*self.A_4*self.core_ci)
    
def make_cicb(self, confidence, draws, hyp, boundary):

    def simulate_Q(i): 
        np.random.seed(i)
        mc = np.sort(np.random.uniform(0, 1, self.sample_size))
        return self.hat_q*(mc-self.u_grid) # this is not uniform
    
    p = Pool(os.cpu_count())
    delta_Qs = np.array(p.map(simulate_Q, range(draws)))
    p.close()
    p.join()
    
    def simulate_q(i): 
        np.random.seed(i)
        mc = np.sort(np.random.uniform(0, 1, self.sample_size))
        mcq = q_smooth(mc, self.kernel, *self.band_options, is_sorted = True, boundary = boundary)
        return (mcq-1) # this is uniform

    p = Pool(os.cpu_count())
    delta_qs = np.array(p.map(simulate_q, range(draws)))
    p.close()
    p.join()
    
    if boundary == 'zero':
        delta_qs[:,-self.trim:] = 0
        delta_qs[:,:self.trim] = 0
        delta_Qs[:,-self.trim:] = 0
        delta_Qs[:,:self.trim] = 0
        
    if hyp == 'twosided':
        def _sup(x):
            return np.max(np.abs(x)[:,self.trim:-self.trim], axis = 1)
        def _perc(x):
            return np.percentile(x, confidence+(100-confidence)/2, axis = 0)
    if hyp == 'onesided':
        def _sup(x):
            return np.max(x[:,self.trim:-self.trim], axis = 1)
        def _perc(x):
            return np.percentile(x, confidence, axis = 0)
        
    delta_ts = np.apply_along_axis(lambda x: total_surplus_from_Q(x, *self.part_options), 1, delta_Qs)
    del(delta_Qs)
       
    add_column(self, '_ts_ci', _perc(delta_ts))
    add_column(self, '_ts_cb', _perc(_sup(delta_ts)))
        
    core_ci = self.hat_q*_perc(delta_qs)
    core_cb = self.hat_q*_perc(_sup(delta_qs))
    del(delta_qs)
    
    add_column(self, '_q_ci', core_ci)
    add_column(self, '_q_cb', core_cb)
    
    add_column(self, '_v_ci', self.A_4*core_ci)
    add_column(self, '_v_cb', self.A_4*core_cb)
    
    add_column(self, '_bs_ci', self.a*self.A_3*self.A_4*core_ci)
    add_column(self, '_bs_cb', self.a*self.A_3*self.A_4*core_cb)
    
    add_column(self, '_rev_ci', self.M*self.a*self.A_3*self.A_4*core_ci)
    add_column(self, '_rev_cb', self.M*self.a*self.A_3*self.A_4*core_cb)



import numpy as np
import pandas as pd
import statsmodels.formula.api as smf
import matplotlib.pyplot as plt
import seaborn as sb
from pylab import rcParams

def plot_ci(self, ax):
    
    ax.plot(self.data._u, self.data._hat_ts, label = 'ts', color = 'green')
    ax.plot(self.data._u, self.data._hat_ts + self.data._ts_ci, color = 'green', linestyle = '--',linewidth = .7)
    ax.plot(self.data._u, self.data._hat_ts - self.data._ts_ci, color = 'green', linestyle = '--',linewidth = .7)
    
    ax.plot(self.data._u, self.M*self.data._hat_bs, label = 'M*bs', color = 'red', linewidth=1)
    ax.plot(self.data._u, self.M*(self.data._bs_ci+self.data._hat_bs), 
            color = 'red', linestyle = '--', linewidth = .7)
    ax.plot(self.data._u, self.M*(-self.data._bs_ci+self.data._hat_bs), 
            color = 'red', linestyle = '--', linewidth = .7)
    
    ax.plot(self.data._u, self.data._hat_rev, label = 'rev', color = 'blue')
    ax.plot(self.data._u, self.data._rev_ci + self.data._hat_rev, 
            color = 'blue', linestyle = '--',linewidth = .7)
    ax.plot(self.data._u, -self.data._rev_ci + self.data._hat_rev, 
            color = 'blue', linestyle = '--',linewidth = .7)
    
    ax.axvline(self.opt_u, linewidth = .7, color = 'black', label = 'optimal exclusion', linestyle = 'dotted')
    
    ax.legend(loc = 'upper right')
    ax.set_xlabel('confidence intervals')
    
def plot_cb(self, ax):
    
    ax.plot(self.data._u, self.data._hat_ts, label = 'ts', color = 'green')
    ax.plot(self.data._u, self.data._hat_ts + self.data._ts_cb, color = 'green', linestyle = '--',linewidth = .7)
    ax.plot(self.data._u, self.data._hat_ts - self.data._ts_cb, color = 'green', linestyle = '--',linewidth = .7)
    
    ax.plot(self.data._u, self.M*self.data._hat_bs, color = 'red', linewidth=1)
    ax.plot(self.data._u, self.M*(self.data._bs_cb+self.data._hat_bs), 
            color = 'red', linestyle = '--',linewidth = .7)
    ax.plot(self.data._u, self.M*(-self.data._bs_cb+self.data._hat_bs), 
            color = 'red', linestyle = '--',linewidth = .7)
    
    ax.plot(self.data._u, self.data._hat_rev, color = 'blue')
    ax.plot(self.data._u, self.data._rev_cb + self.data._hat_rev, 
            color = 'blue', linestyle = '--',linewidth = .7)
    ax.plot(self.data._u, -self.data._rev_cb + self.data._hat_rev, 
            color = 'blue', linestyle = '--',linewidth = .7)
    
    ax.axvline(self.opt_u, linewidth = .7, color = 'black', linestyle = 'dotted')
    
    ax.set_xlabel('confidence bands')

def plot_counterfactuals(self):
    rcParams['figure.figsize'] = 7, 7/3
    fig, (ax1, ax2) = plt.subplots(1,2, sharey = True)
    
    plot_ci(self, ax1)
    plot_cb(self, ax2)
    
    plt.tight_layout()
    plt.show()

def plot_bidders(self, ax):
    
    sb.countplot(x = self.data.groupby(by = 'auctionid')._bidders.first().astype(int), 
                 facecolor=(0, 0, 0, 0),
                 linewidth=1,
                 edgecolor='black', 
                 ax = ax)
    
    ax.set_xlabel('bidders')
    
def plot_bid_residuals(self, ax):
    
    sb.histplot(data = self.data.loc[self.active_index,'_resid'], 
                stat = 'density', 
                bins = 50, 
                facecolor=(0, 0, 0, 0),
                linewidth=1,
                edgecolor='black', 
                ax = ax);
    
    ax.set_xlabel('bid residuals')
    ax.set_ylabel('density')
    
def plot_aux(self, ax):
    
    ax.plot(self.u_grid, self.A_1, label = '$A_1$')
    ax.plot(self.u_grid, self.A_2, label = '$A_2$')
    ax.plot(self.u_grid, self.A_3, label = '$A_3$')
    ax.plot(self.u_grid, self.A_4, label = '$A_4$')
    ax.set_xlabel('auxilliary functions')
    ax.legend()
    
def plot_densities(self, ax):
    
    ciq = self.ci_two*self.hat_q/np.sqrt(self.sample_size*self.band)
    
    ax.plot(self.u_grid, 
             self.hat_q, 
             label = 'smooth $\hat q(u)$', linewidth = .7, color = 'blue')
    ax.plot(self.u_grid, self.hat_q+ciq, linestyle = '--', linewidth = .7, color = 'blue')
    ax.plot(self.u_grid, self.hat_q-ciq, linestyle = '--', linewidth = .7, color = 'blue')
    
    
    ax.plot(self.u_grid, 
             self.hat_f*self.scale, 
             color = 'red', 
             label = 'smooth $\hat f(b)$ (scale matched)', 
             linewidth=1)
    
    cif = self.ci_two*np.sqrt(self.hat_f)/np.sqrt(self.sample_size*self.band)
    
    ax.plot(self.u_grid, 
             (self.hat_f+cif)*self.scale, 
             color = 'red', 
             linewidth=1, linestyle = '--')
    
    ax.plot(self.u_grid, 
             (self.hat_f-cif)*self.scale, 
             color = 'red', 
             linewidth=1, linestyle = '--')
    
    ax.legend()
    
def plot_quantiles(self, ax):
    
    avg_fitted = 1 #self.data._fitted.mean()

    if self.model_type == 'multiplicative':
        b_qf = self.hat_Q * avg_fitted
        v_qf = self.hat_v * avg_fitted

    if self.model_type == 'additive':
        b_qf = self.hat_Q + avg_fitted
        v_qf = self.hat_v + avg_fitted

    ax.plot(self.u_grid, b_qf, label = 'bid quantile function')
    ax.plot(self.u_grid, v_qf, label = 'value quantile function')
    ax.legend()
    
def plot_val_residuals(self, ax):
    sb.histplot(data = self.data._latent_resid, 
                stat = 'density', 
                bins = 50, 
                facecolor=(0, 0, 0, 0),
                linewidth=1,
                edgecolor='black', 
                ax = ax);
    
    ax.set_xlabel('value residuals')
    ax.set_ylabel('')
    
def plot_stats(self):
    rcParams['figure.figsize'] = 7, 7
    fig, ((ax1, ax3), (ax2, ax6), (ax5, ax4)) = plt.subplots(3,2)

    plot_bidders(self, ax1)
    plot_bid_residuals(self, ax2)
    plot_aux(self, ax3)
    plot_densities(self, ax4)
    plot_quantiles(self, ax5)
    plot_val_residuals(self, ax6)

    plt.tight_layout()
    plt.show()


